import gym
import numpy
from numpy import random
from numba import njit


@njit(cache=True)
def update_value_matrix(r_mat, t_mat, gamma, vm, max_iteration=-1):
    diff = 1.0
    cur_vm = numpy.copy(vm)
    ns, na = r_mat.shape
    iteration = 0

    while diff > 1.0e-4 and (
            (max_iteration < 0) or 
            (max_iteration > iteration and max_iteration > 1) or
            (iteration < 1 and random.random() < max_iteration)):
        iteration += 1
        old_vm = numpy.copy(cur_vm)
        for s in range(ns):
            for a in range(na):
                exp_q = 0.0
                for sn in range(ns):
                    exp_q += t_mat[s,a,sn] * numpy.max(cur_vm[sn])
                cur_vm[s,a] = r_mat[s,a] + gamma * exp_q
        diff = numpy.sqrt(numpy.mean((old_vm - cur_vm)**2))
    return cur_vm

class AnyMDPSolverOpt(object):
    """
    Solver for AnyMDPEnv with Bellman Equation and Value Iteration
    Suppose to know the ground truth of the environment
    """
    def __init__(self, env, gamma=0.98):
        if(not env.task_set):
            raise Exception("AnyMDPEnv is not initialized by 'set_task', must call set_task first")
        self.n_actions = env.action_space.n
        self.n_states = env.observation_space.n
        self.transition_matrix = env.transition_matrix
        self.reward_matrix = env.reward_matrix
        self.value_matrix = numpy.zeros((self.n_states, self.n_actions))
        self.gamma = gamma
        self.q_solver()

    def q_solver(self, gamma=0.99):
        self.value_matrix = update_value_matrix(self.reward_matrix, self.transition_matrix, gamma, self.value_matrix)

    def policy(self, state):
        return numpy.argmax(self.value_matrix[state])


"""
Gym Environment For Maze3D
"""
import numpy
import gym
import pygame
from numpy import random

from gym import error, spaces, utils
from gym.utils import seeding

class AnyMDPEnv(gym.Env):
    def __init__(self, max_steps:int=5000):
        self.max_steps = max_steps
        self.observation_space = spaces.Discrete(2)
        self.action_space = spaces.Discrete(2)
        self.task_set = False

    def set_task(self, task_config):
        self.transition_matrix = task_config["transition"]
        self.reward_matrix = task_config["reward"]
        self.reward_noise = task_config["reward_noise"]
        self.reward_noise_type = task_config["reward_noise_type"]

        ns1, na1, ns2 = self.transition_matrix.shape
        ns3, na2 = self.reward_matrix.shape

        assert ns1 == ns2 == ns3, "Transition matrix and reward matrix must have the same number of states"
        assert na1 == na2, "Transition matrix and reward matrix must have the same number of actions"
        assert ns1 > 0, "State space must be at least 1"
        assert na1 > 1, "Action space must be at least 2"

        self.observation_space = spaces.Discrete(ns1)
        self.action_space = spaces.Discrete(na1)

        self.n_states = ns1
        self.n_actions = na1
        self.task_set = True
        self.need_reset = True

    def reset(self):
        if(not self.task_set):
            raise Exception("Must call \"set_task\" first")
        
        self.steps = 0
        self.need_reset = False
        self._state = random.randint(0, self.n_states)
        return self._state, {"steps": self.steps}

    def step(self, action):
        if(self.need_reset or not self.task_set):
            raise Exception("Must \"set_task\" and \"reset\" before doing any actions")
        assert action < self.n_actions, "Action must be less than the number of actions"
        transition_gt = self.transition_matrix[self._state, action]
        reward_gt = self.reward_matrix[self._state, action]
        reward_gt_noise = self.reward_noise[self._state, action]
        next_state = random.choice(self.n_states, p=transition_gt)
        if(self.reward_noise_type == 'normal'):
            reward = random.normal(reward_gt, reward_gt_noise)
        elif(self.reward_noise_type == 'binomial'):
            if(reward_gt > 0):
                reward = random.binomial(1, reward_gt)
            else:
                reward = - random.binomial(1, abs(reward_gt))

        info = {"steps": self.steps, "transition_gt": transition_gt, "reward_gt": reward_gt}
        self.steps += 1
        self._state = next_state
        done = (self.steps >= self.max_steps)
        if(done):
            self.need_reset = True

        return next_state, reward, done, info



#   Copyright (c) 2021 DeepEvolution Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import _io
import numpy
import gym
from numpy import random


def TaskSamplerV2(seed=None,
                n_emb=16,
                n_hidden=64,
                n_vocab=256,
                n_gram=3,
                _lambda=5.0):
    if(seed is not None):
        numpy.random.seed(seed)
    if(isinstance(n_gram, list)):
        n_gram = random.choice(n_gram)
    word_emb = numpy.random.normal(0, 1.0, size=(n_vocab, n_emb))
    weights_inputlayer = numpy.random.normal(0, 1.0, size=(n_gram, n_emb, n_hidden))
    bias_inputlayer = numpy.random.normal(0, 1.0, size=(n_gram, 1, n_hidden))
    weights_outputlayer = numpy.random.normal(0, 1.0, size=(n_hidden, n_vocab))
    bias_outputlayer = numpy.random.normal(0, 1.0, size=(1, n_vocab))
    return {
                'word_emb': word_emb,
                'weights_inputlayer': weights_inputlayer,
                'bias_inputlayer': bias_inputlayer,
                'weights_outputlayer': weights_outputlayer,
                'bias_outputlayer': bias_outputlayer,
                '_lambda': _lambda,
                'n_emb': n_emb,
                'n_hidden': n_hidden,
                'n_vocab': n_vocab,
                'n_gram': n_gram
            }

def TaskSamplerV1(seed=None,
                n_vocab=64, 
                n_patterns=10, 
                n_gram=64,
                error_ratio=0.1):
    patterns = []
    if(seed is not None):
        numpy.random.seed(seed)
    if(isinstance(n_gram, list)):
        n_gram = random.choice(n_gram)
    for _ in range(n_patterns):
        l_r = max(3, numpy.random.poisson(n_gram))
        patterns.append(random.randint(0, n_vocab, size=(l_r), dtype="int32"))
    return {
            'patterns': patterns,
            'n_vocab': n_vocab,
            'n_patterns': n_patterns,
            'error_ratio': error_ratio,
            'n_gram': n_gram
        }



import numpy
from numpy import random
from attr import attrs, attrib
from l3c.mazeworld.envs.utils import conv2d_numpy

"""
Generating Topology For A Maze
"""

@attrs
class Rectangle(object):
    lb = attrib(type=list, default=None)
    rt = attrib(type=list, default=None)

    def resample(self, cells, max_size=10, min_size=2):
        m_x, m_y = cells.shape

        # Sample the rectangle width and length first
        w_x = random.randint(min_size, max_size + 1)
        w_y = random.randint(min_size, max_size + 1)

        kernel = numpy.ones((w_x, w_y), dtype="float32")

        # Use convolution to make sure no overlapping in rectangles
        overlap = conv2d_numpy(cells, kernel, stride=(1, 1))
        freerows, freecols = numpy.where(overlap < 0.5)

        if(freerows.shape[0] > 0):
            sel_idx = random.randint(0, freerows.shape[0])
        else:
            return False

        self.lb = [freerows[sel_idx], freecols[sel_idx]]
        self.rt = [self.lb[0] + w_x - 1, self.lb[1] + w_y - 1]
        return True

    def refresh_rectangle(self, cells):
        m_x, m_y = cells.shape
        b_x = max(0, self.lb[0] - 1)
        b_y = max(0, self.lb[1] - 1)
        e_x = min(m_x, self.rt[0] + 2)
        e_y = min(m_y, self.rt[1] + 2)
        cells[b_x:e_x, b_y:e_y] = 1

    def refresh_occupancy(self, cells):
        self.refresh_rectangle(cells)
        cells[self.lb[0]:(self.rt[0] + 1), self.lb[1]:(self.rt[1] + 1)] = 0

def genmaze_largeroom(n, room_number, room_size=(2,4)):
    cells_occ = numpy.zeros((n-2, n-2), dtype=numpy.int8)
    cells_wall = numpy.ones((n-2, n-2), dtype=numpy.int8)
    rects = []

    # TRY PUT 6 RECTANGLES
    max_try = 5
    for _ in range(room_number):
        rect = Rectangle()
        for _ in range(max_try):
            is_succ = rect.resample(cells_occ, min_size=room_size[0], max_size=room_size[1])
            if(is_succ):
                rect.refresh_rectangle(cells_occ)
                rects.append(rect)
                break
    for rect in rects:
        rect.refresh_occupancy(cells_wall)

    cell_occs = numpy.ones((n, n), dtype=numpy.int8)
    cell_walls = numpy.ones((n, n), dtype=numpy.int8)
    cell_occs[1:n-1, 1:n-1] = cells_occ
    cell_walls[1:n-1, 1:n-1] = cells_wall

    return cell_occs, cell_walls, rects

def genmaze_by_primwall(n, allow_loops=True, wall_density=0.30):
    # Dig big rooms in the region
    cell_occs, cell_walls, rects = genmaze_largeroom(n, random.randint(0, (n - 2) ** 2 // 16))

    # Dig the initial holes
    for i in range(1, n, 2):
        for j in range(1, n, 2):
            if(not cell_occs[i,j]):
                cell_walls[i,j] = 0

    #Initialize the logics for prim based maze generation
    wall_dict = dict()
    path_dict = dict()
    rev_path_dict = dict()
    path_idx = 0
    for i in range(1, n - 1):
        for j in range(1, n - 1):
            if(cell_walls[i,j]): # we will keep the axial point
                wall_dict[i, j] = 0
            elif(not cell_occs[i,j]):
                path_dict[i,j] = path_idx
                rev_path_dict[path_idx] = [(i,j)]
                path_idx += 1
    for rect in rects:
        xb = rect.lb[0] + 1
        yb = rect.lb[1] + 1
        xe = rect.rt[0] + 2
        ye = rect.rt[1] + 2
        rev_path_dict[path_idx] = []
        for i in range(xb, xe):
            for j in range(yb, ye):
                path_dict[i,j] = path_idx
                rev_path_dict[path_idx].append((i, j))
        path_idx += 1

    #Prim the wall until all points are connected
    max_cell_walls = numpy.prod(cell_walls[1:-1, 1:-1].shape)
    while len(rev_path_dict) > 1 or (allow_loops and numpy.sum(cell_walls[1:-1, 1:-1]) > max_cell_walls * wall_density):
        wall_list = list(wall_dict.keys())
        random.shuffle(wall_list)
        for i, j in wall_list:
            new_path_id = -1
            connected_path_id = dict()
            abandon_path_id = dict()
            max_duplicate = 1

            for d_i, d_j in [(i - 1, j), (i + 1, j), (i, j - 1), (i, j + 1)]:
                if((d_i > 0  and d_i < n and d_j > 0 and d_j < n)
                        and cell_walls[d_i, d_j] < 1):
                    # calculate duplicate path id that might creat a loop
                    if path_dict[d_i, d_j] not in connected_path_id:
                        connected_path_id[path_dict[d_i, d_j]] = 1
                    else:
                        connected_path_id[path_dict[d_i, d_j]] += 1
                    if(connected_path_id[path_dict[d_i, d_j]] > max_duplicate):
                        max_duplicate = connected_path_id[path_dict[d_i, d_j]]

                    # decide the new path_id and find those to be deleted
                    if(path_dict[d_i, d_j] < new_path_id or new_path_id < 0):
                        if(new_path_id >= 0):
                            abandon_path_id[new_path_id] = (new_i, new_j)
                        new_path_id = path_dict[d_i, d_j]
                        new_i = d_i
                        new_j = d_j
                    elif(path_dict[d_i, d_j] != new_path_id): # need to be abandoned
                        abandon_path_id[path_dict[d_i, d_j]] = (d_i, d_j)
            if(len(abandon_path_id) >= 1 and max_duplicate < 2):
                break
            if(len(abandon_path_id) >= 1 and max_duplicate > 1 and allow_loops):
                break
            if(allow_loops and len(rev_path_dict) < 2 and random.random() < 0.2):
                break

        if(new_path_id < 0):
            continue

        # add the released wall
        rev_path_dict[new_path_id].append((i,j))
        path_dict[i,j] = new_path_id
        cell_walls[i,j] = 0
        del wall_dict[i,j]

        # merge the path
        for path_id in abandon_path_id:
            rev_path_dict[new_path_id].extend(rev_path_dict[path_id])
            for t_i_o, t_j_o in rev_path_dict[path_id]:
                path_dict[t_i_o,t_j_o] = new_path_id
            del rev_path_dict[path_id]
    return cell_walls

if __name__=="__main__":
    print(genmaze_by_primwall(15).astype("int8"))



import gym
import numpy
from numpy import random
from l3c.anymdp.anymdp_solver_opt import update_value_matrix


class AnyMDPSolverMBRL(object):
    """
    Solver for AnyMDPEnv with Model-based Reinforcement Learning
    """
    def __init__(self, env, gamma=0.98, c=0.02):
        """
        The constructor for the class AnyMDPSolverQ
        The exploration strategy is controlled by UCB-H with c as its hyperparameter. Increasing c will encourage exploration
        Simulation of the ideal policy when the ground truth is not known
        """
        if(not env.task_set):
            raise Exception("AnyMDPEnv is not initialized by 'set_task', must call set_task first")
        self.n_actions = env.action_space.n
        self.n_states = env.observation_space.n
        self.transition_matrix = env.transition_matrix
        self.reward_matrix = env.reward_matrix
        self.value_matrix = numpy.zeros((self.n_states, self.n_actions))
        self.est_r = numpy.zeros((self.n_states, self.n_actions))
        self.est_r_cnt = numpy.zeros((self.n_states, self.n_actions))
        self.est_t = numpy.ones((self.n_states, self.n_actions, self.n_states))
        self.gamma = gamma
        self._c = c / (1.0 - self.gamma)
        self.step = 0


    def learner(self, s, a, ns, r, done):
        # Update the environment model estimation
        if(self.est_r_cnt[s,a] < 1):
            self.est_r[s,a] = r
        else:
            self.est_r[s,a] += (r - self.est_r[s,a])/self.est_r_cnt[s,a]
        self.est_r_cnt[s,a]+= 1
        self.est_t[s,a,ns] += 1
        est_t = self.est_t / numpy.sum(self.est_t, axis=-1, keepdims=True)

        # Update the value matrix
        self.value_matrix = update_value_matrix(self.est_r, est_t, self.gamma, self.value_matrix, max_iteration=0.2)
        self.step += 1

    def policy(self, state):
        # Apply UCB-H exploration strategy
        values = self._c * numpy.sqrt(numpy.log(self.step + 1) / numpy.clip(self.est_r_cnt[state], 1.0, 1.0e+8))  + self.value_matrix[state]
        return numpy.argmax(values)


if __name__=="__main__":
    import gym
    from l3c.anymdp import AnyMDPEnv, AnyMDPSolverOpt, AnyMDPSolverQ, AnyMDPSolverMBRL, AnyMDPTaskSampler, Resampler

    env = gym.make("anymdp-v0", max_steps=32000)
    task = AnyMDPTaskSampler(128, 5)
    prt_freq = 1000
    env.set_task(Resampler(task))
    print(task)

    # Test Random Policy
    state, info = env.reset()
    acc_reward = 0
    epoch_reward = 0
    done = False

    while not done:
        action = env.action_space.sample()
        state, reward, done, info = env.step(action)
        acc_reward += reward
        epoch_reward += reward
        if(info["steps"] % prt_freq == 0 and info["steps"] > 0):
            print("Step:{}\tEpoch Reward: {}".format(info["steps"], epoch_reward))
            epoch_reward = 0
    print("Random Policy Summary: {}".format(acc_reward))

    # Test AnyMDPSolverOpt
    solver = AnyMDPSolverOpt(env)
    state, info = env.reset()
    done = False
    acc_reward = 0
    epoch_reward = 0

    while not done:
        action = solver.policy(state)
        state, reward, done, info = env.step(action)
        acc_reward += reward
        epoch_reward += reward
        if(info["steps"] % prt_freq == 0 and info["steps"] > 0):
            print("Step:{}\tEpoch Reward: {}".format(info["steps"], epoch_reward))
            epoch_reward = 0
    print("Optimal Solver Summary:  {}".format(acc_reward))

    # Test AnyMDPSolverQ
    solver = AnyMDPSolverQ(env)
    state, info = env.reset()
    done = False
    acc_reward = 0
    epoch_reward = 0

    while not done:
        action = solver.policy(state)
        next_state, reward, done, info = env.step(action)
        solver.learner(state, action, next_state, reward, done)
        acc_reward += reward
        epoch_reward += reward
        state = next_state
        if(info["steps"] % prt_freq == 0 and info["steps"] > 0):
            print("Step:{}\tEpoch Reward: {}".format(info["steps"], epoch_reward))
            epoch_reward = 0
    print("Q Solver Summary: {}".format(acc_reward))

    # Test AnyMDPSolverMBRL
    solver = AnyMDPSolverMBRL(env)
    state, info = env.reset()
    done = False
    acc_reward = 0
    epoch_reward = 0

    while not done:
        action = solver.policy(state)
        next_state, reward, done, info = env.step(action)
        solver.learner(state, action, next_state, reward, done)
        acc_reward += reward
        epoch_reward += reward
        state = next_state
        if(info["steps"] % prt_freq == 0 and info["steps"] > 0):
            print("Step:{}\tEpoch Reward: {}".format(info["steps"], epoch_reward))
            epoch_reward = 0
    print("MBRL Solver Summary: {}".format(acc_reward))

    print("Test Passed")


"""
Core File of Maze Env
"""
import os
import numpy
import pygame
from numpy import random
from collections import namedtuple
from numpy import random as npyrnd
from numpy.linalg import norm
from copy import deepcopy
from l3c.mazeworld.envs.grid_ops import genmaze_by_primwall


def gentext(cell_walls, textlib_walls, textlib_grounds, textlib_ceilings):
    n = cell_walls.shape[0]
    cell_texts = numpy.random.randint(0, len(textlib_walls), size=cell_walls.shape)

    #Paint the texture of passways to ground textures 
    for i in range(1, n - 1):
        for j in range(1, n - 1):
            if(cell_walls[i,j] < 1):
                cell_texts[i,j] = 0

    text_ground = random.randint(0, len(textlib_grounds))
    text_ceiling = random.randint(0, len(textlib_ceilings))

    return cell_texts, text_ground, text_ceiling

def idx_trans(idx, n):
    return (idx // n, idx % n)

def gentargets(cell_walls, landmarks_number):
    #Randomize a start point and n landmarks
    n = cell_walls.shape[0]
    landmarks_likelihood = numpy.random.rand(*cell_walls.shape) - cell_walls
    idxes = numpy.argsort(landmarks_likelihood, axis=None)
    topk_idxes = idxes[-landmarks_number:]
    landmarks = [idx_trans(i, n) for i in idxes[-landmarks_number:]]

    cell_landmarks = numpy.zeros_like(cell_walls, dtype=numpy.int8) - 1
    for i,idx in enumerate(landmarks):
        cell_landmarks[tuple(idx)] = int(i)
    cell_landmarks = cell_landmarks.astype(cell_walls.dtype)
    return landmarks, cell_landmarks

def genstart(cell_walls, cell_landmarks):
    n = cell_walls.shape[0]
    landmarks_likelihood = numpy.random.rand(n, n) - cell_walls - cell_landmarks
    idxes = numpy.argsort(landmarks_likelihood, axis=None)
    return idx_trans(idxes[-1], n)

def dict_formatter(data):
    str=""
    for key, value in data.items():
        str += f"\n\n{key}:\n{value}\n\n"
    return str

class MazeTaskManager(object):
    def __init__(self, texture_dir, verbose=False):
        pathes = os.path.split(os.path.abspath(__file__))
        texture_dir = os.sep.join([pathes[0], texture_dir])
        texture_files = os.listdir(texture_dir)
        texture_files.sort()
        textlib_grounds = []
        textlib_ceilings = []
        textlib_walls = []
        for file_name in texture_files:
            if(file_name.find("wall") == 0):
                textlib_walls.append(pygame.surfarray.array3d(pygame.image.load(os.sep.join([texture_dir, file_name]))))
            if(file_name.find("ground") == 0):
                textlib_grounds.append(pygame.surfarray.array3d(pygame.image.load(os.sep.join([texture_dir, file_name]))))
            if(file_name.find("ceiling") == 0):
                textlib_ceilings.append(pygame.surfarray.array3d(pygame.image.load(os.sep.join([texture_dir, file_name]))))
        self.textlib_walls = numpy.asarray(textlib_walls, dtype="float32")
        self.textlib_grounds = numpy.asarray(textlib_grounds, dtype="float32")
        self.textlib_ceilings = numpy.asarray(textlib_ceilings, dtype="float32")
        self.verbose = verbose

    @property
    def n_texts(self):
        return self.grounds.shape[0]

    def sample_cmds(self, n, commands_sequence):
        xs = numpy.random.randint(0, n, commands_sequence)
        for i in range(xs.shape[0]):
            if(i > 0):
                if(xs[i] == xs[i-1]):
                    xs[i] = (xs[i] + random.randint(1, n)) % n
        return xs

    def sample_task(self,
            n_range:tuple=(9, 25), 
            allow_loops:bool=True, 
            cell_size_range=(1.5, 4.5), 
            wall_height_range=(2.0, 6.0), 
            agent_height_range=(1.6, 2.0),
            wall_density_range=(0.2, 0.4),
            landmarks_number_range=(5, 15),
            commands_sequence=200,
            step_reward=0.0,
            collision_reward=-0.20,
            goal_reward=None,
            seed=None,
            verbose=False):
        # Initialize the maze ...
        if(seed is not None):
            seed = time.time() * 1000 % 65536
        numpy.random.seed(seed)

        cell_size = random.uniform(*cell_size_range)
        wall_height = random.uniform(*wall_height_range)
        agent_height = random.uniform(*agent_height_range)
        wall_density = random.uniform(*wall_density_range)
        landmarks_number = random.randint(*landmarks_number_range)
        n = random.randint(*n_range)
        if(n % 2 == 0):
            n += 1
        assert n > 6, "Minimum required cells are 7"
        assert landmarks_number > 1, "There must be at least 1 goal, thus landmarks_number must > 1"
        if(landmarks_number > 15):
            landmarks_number = 15
            print("landmarks number too much, set to 15")
        if(self.verbose):
            print("Generating an random maze of size %dx%d, with allow loops=%s, crowd ratio=%f"%(n, n, allow_loops, wall_density))

        # Generate the wall topology
        cell_walls = genmaze_by_primwall(n, allow_loops=allow_loops, wall_density=wall_density)

        # Selects the texture
        cell_texts, ground_text, ceiling_text = gentext(cell_walls, self.textlib_walls, self.textlib_grounds, self.textlib_ceilings)

        # Generate landmarks (Potential Navigation Targets)
        landmarks, cell_landmarks = gentargets(cell_walls, landmarks_number)

        # Generate start location
        start = genstart(cell_walls, cell_landmarks)

        #Calculate goal reward
        if(goal_reward is None):
            def_goal_reward = n * numpy.sqrt(n) / 60.0
        else:
            def_goal_reward = goal_reward
        assert def_goal_reward > 0, "goal reward must be > 0"

        # Sample Commands Sequences
        commands_sequence = self.sample_cmds(len(landmarks), commands_sequence)

        task = {"start":start,
                "cell_walls":cell_walls,
                "cell_texts":cell_texts,
                "cell_size":cell_size,
                "ground_text":ground_text,
                "ceiling_text":ceiling_text,
                "step_reward":step_reward,
                "goal_reward":def_goal_reward,
                "collision_reward":collision_reward,
                "wall_height":wall_height,
                "agent_height":agent_height,
                "commands_sequence":commands_sequence,
                "landmarks_coordinates":landmarks,
                "cell_landmarks":cell_landmarks}
        
        if(verbose):
            print("\n\n-------Successfully Generate Task-------\n")
            print(dict_formatter(task))
            print("\n----------------------------------------\n\n")

        return task

    def resample_task(self, task, 
            resample_cmd=True, 
            resample_start=True, 
            resample_landmarks=False,
            resample_landmarks_color=False,
            seed=None, verbose=False):
        # Randomize a start point and n landmarks while keeping the scenario still
        if(seed is not None):
            seed = time.time() * 1000 % 65536
        numpy.random.seed(seed)
        n = task["cell_walls"].shape[0]

        def idx_trans(idx):
            return (idx // n, idx % n)

        landmarks_number = len(task["landmarks_coordinates"])
        if(resample_landmarks):
            landmarks, cell_landmarks = gentargets(task["cell_walls"], landmarks_number)
        elif(resample_landmarks_color):
            landmarks = deepcopy(task["landmarks_coordinates"])
            random.shuffle(landmarks)
            cell_landmarks = numpy.zeros_like(task["cell_walls"]) - 1
            for i,idx in enumerate(landmarks):
                cell_landmarks[tuple(idx)] = int(i)
            cell_landmarks = cell_landmarks.astype(task["cell_walls"].dtype)
        else:
            landmarks = deepcopy(task["landmarks_coordinates"])
            cell_landmarks = deepcopy(task["cell_landmarks"])

        # Generate start location
        start = genstart(task["cell_walls"], cell_landmarks)

        # Generate command sequences
        commands_sequence = self.sample_cmds(len(landmarks), len(task["commands_sequence"]))

        new_task = deepcopy(task)
        new_task["start"] = start
        new_task["landmarks_coordinates"] = landmarks
        new_task["cell_landmarks"] = cell_landmarks
        new_task["commands_sequence"] = commands_sequence

        if(verbose):
            print("\n\n-------Successfully Resample Task-------\n")
            print(dict_formatter(task))
            print("\n----------------------------------------\n\n")

        return new_task

MAZE_TASK_MANAGER=MazeTaskManager("img")
MazeTaskSampler = MAZE_TASK_MANAGER.sample_task
Resampler = MAZE_TASK_MANAGER.resample_task


if __name__=="__main__":
    task = MazeTaskSampler(verbose=False)
    print(task)
    print(Resampler(task))
    print(Resampler(task, resample_landmarks_color=True))
    print(Resampler(task, resample_landmarks=True))
